/*
******************************************************************************

   LibBUF: plain functions

   Copyright (C) 2001 Christoph Egger   [Christoph_Egger@t-online.de]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************
*/


#include <ggi/internal/buf.h>
#include <ggi/internal/buf_debug.h>
#include <ggi/internal/galloc_ops.h>



int ggiBufClad(ggiBuf_t buf, int numbufs, ggiBuf_t *bufs, ggiBuf_t *res)
{
        ggiGA_resource_handle *buf_ga;
        ggi_visual *vis;

        if (buf == NULL) return GGI_EARGINVAL;

        buf_ga = GA_LL(buf);
        if (buf_ga == NULL) return GGI_EARGINVAL;

	vis = NULL;
        while (*buf_ga) {
	  if ((ggiGAGetType(*buf_ga) & GA_RT_TYPE_MASK) == GA_RT_BUFFER) 
	    vis = ggiGAGetVisual(*buf_ga);
	  if (vis != NULL) break;
          buf_ga++;
        }
        if (vis == NULL) return GGI_EARGINVAL;

	return ggiBuf_clad(vis, 0, numbufs, bufs, res);
}	/* ggiBufClad */

int ggiBufCladVis(ggi_visual_t vis, int numbufs, ggiBuf_t *bufs)
{
	return ggiBuf_clad(vis, BUF_CLAD_VISUAL, numbufs, bufs, NULL);
}	/* ggiBufCladVis */

int ggiBufSetGCForeground(ggiBuf_t buf, ggi_bufval val)
{
	ggi_ll_fgbg_gc_t gc;

	if (buf == NULL) return GGI_EARGINVAL;
	gc = GGI_LL_GC(buf);
	if (gc == NULL) return GGI_EARGINVAL;
	gc->fgsrc = val;
	gc = GGI_LL_GC2(buf);
	if (gc == NULL) return GGI_EARGINVAL;
	gc->fgdest = val;

	return 0;
}	/* ggiBufSetGCForeground */

int ggiBufSetGCBackground(ggiBuf_t buf, ggi_bufval val)
{
	ggi_ll_fgbg_gc_t gc;

	if (buf == NULL) return GGI_EARGINVAL;
	gc = GGI_LL_GC(buf);
	if (gc == NULL) return GGI_EARGINVAL;
	gc->bgsrc = val;
	gc = GGI_LL_GC2(buf);
	if (gc == NULL) return GGI_EARGINVAL;
	gc->bgdest = val;

	return 0;
}	/* ggiBufSetGCBackground */

int ggiBufSetGCFlags(ggiBuf_t buf, enum ggiBuf_flags flags)
{
        ggiGA_resource_handle *buf_ga, carb_h, tank_h;
	struct ggiGA_resource_props *carb_p, *tank_p;
	ggi_ll_math_gc_t mathgc;
	enum ggi_ll_math mathmask;

        if (buf == NULL) goto inval;

        buf_ga = GA_LL(buf);
        if (buf_ga == NULL) goto inval;

	tank_h = NULL;
	carb_h = NULL;

        while (*buf_ga) {
          if (ggiGAIsCarb(*buf_ga)) carb_h = *buf_ga;
	  if ((ggiGAGetType(*buf_ga) & GA_RT_TYPE_MASK) == GA_RT_BUFFER)
	    if (ggiGAIsTank(*buf_ga)) tank_h = *buf_ga;
          buf_ga++;
        }

	if (tank_h == NULL) goto inval;
	if (carb_h == NULL) goto inval;
	carb_p = ggiGAGetProperties(carb_h);
	LIB_ASSERT(carb_p != NULL);
	tank_p = ggiGAGetProperties(tank_h);

	mathgc = GGI_LL_MATHGC(buf);
	if (mathgc == NULL) goto inval;
	mathmask = 0;

	/* Caveat: below we take advantage that the LL math enum and the 
	 * LibBuf flags are mostly identical.  If this fact changes, we
	 * will need to do convertions here.
	 */
	switch(ggiGAGetType(tank_h)) {
	case GA_RT_BUFFER_ZBUFFER:
	  mathmask = BUF_CMP_MASK | BUF_CMP_SIGNED;
	  if (((~(carb_p->sub.carb.can.math)) & flags) & mathmask) 
	    goto badflag;
	  if ((carb_p->sub.carb.must.math & (~flags)) & mathmask) goto badflag;
	  if ((flags & BUF_DISPOSE) &&
	      !(carb_p->sub.carb.can.math & LL_MATH_Z_DISPOSE))
	    goto badflag;
	  if (!(flags & BUF_DISPOSE) &&
	      (carb_p->sub.carb.must.math & LL_MATH_Z_DISPOSE))
	    goto badflag;
	  if ((flags & (BUF_DISPOSE << BUF_TOGGLE_SHIFT)) &&
	      !(carb_p->sub.carb.can.math & 
	       (LL_MATH_Z_DISPOSE << LL_MATH_ALTSHIFT)))
	    goto badflag;
	  if (!(flags & (BUF_DISPOSE << BUF_TOGGLE_SHIFT)) &&
	      (carb_p->sub.carb.must.math & 
	       (LL_MATH_Z_DISPOSE << LL_MATH_ALTSHIFT)))
	    goto badflag;
	  mathgc->math = (flags & mathmask) | carb_p->sub.carb.must.math;
	  break;
	case GA_RT_BUFFER_ABUFFER:
	  mathmask = BUF_BLENDTYPE_MASK | BUF_DISPOSE | BUF_BLEND_POSTDIV | 
	    ((BUF_BLENDTYPE_MASK | BUF_DISPOSE | BUF_BLEND_POSTDIV) 
	     << BUF_ALTBLEND_SHIFT);
	  if (((~(carb_p->sub.carb.can.math)) & flags) & mathmask) 
	    goto badflag;
	  if ((carb_p->sub.carb.must.math & (~flags)) & mathmask) goto badflag;
	  mathgc->math = (flags & mathmask) | carb_p->sub.carb.must.math;
	  mathgc->math |= (flags & BUF_BLEND_ALT);
	  break;
	default:
	  BUFDPRINT("Buffer not Alpha or Z, setflags not implemented yet.\n");
	  return GGI_EARGINVAL;
	  break;
	}
	return 0;
 badflag:
	BUFDPRINT("Requested flag not supported by requested object\n");
	return GGI_EARGINVAL;
 inval:
	BUFDPRINT("Fishy buf_t in ggiBufSetGCFlags\n");
	return GGI_EARGINVAL;
}	/* ggiBufSetGCFlags */

int ggiBufGetGCForeground(ggiBuf_t buf, ggi_bufval *val)
{
	ggi_ll_fgbg_gc_t gc;
	if (buf == NULL) return GGI_EARGINVAL;

	if (buf == NULL) return GGI_EARGINVAL;
	gc = GGI_LL_GC(buf);
	if (gc == NULL) return GGI_EARGINVAL;

	if (val == NULL) return GGI_EARGINVAL;

	*val = gc->fgsrc;
	return 0;
}	/* ggiBufGetGCForeground */

int ggiBufGetGCBackground(ggiBuf_t buf, ggi_bufval *val)
{
	ggi_ll_fgbg_gc_t gc;
	if (buf == NULL) return GGI_EARGINVAL;

	if (buf == NULL) return GGI_EARGINVAL;
	gc = GGI_LL_GC(buf);
	if (gc == NULL) return GGI_EARGINVAL;

	if (val == NULL) return GGI_EARGINVAL;

	*val = gc->bgsrc;
	return 0;
}	/* ggiBufGetGCForeground */

int ggiBufGetGCFlags(ggiBuf_t buf, enum ggiBuf_flags *flags)
{
        ggiGA_resource_handle *buf_ga, carb_h, tank_h;
	struct ggiGA_resource_props *carb_p, *tank_p;
	ggi_ll_math_gc_t mathgc;
	enum ggi_ll_math mathmask;

        if (buf == NULL) goto inval;

        buf_ga = GA_LL(buf);
        if (buf_ga == NULL) goto inval;

	tank_h = NULL;
	carb_h = NULL;

        while (*buf_ga) {
          if (ggiGAIsCarb(*buf_ga)) carb_h = *buf_ga;
	  if ((ggiGAGetType(*buf_ga) & GA_RT_TYPE_MASK) == GA_RT_BUFFER) 
	    if (ggiGAIsTank(*buf_ga)) tank_h = *buf_ga;
          buf_ga++;
        }

	if (tank_h == NULL) goto inval;
	if (carb_h == NULL) goto inval;
	carb_p = ggiGAGetProperties(carb_h);
	tank_p = ggiGAGetProperties(tank_h);

	mathgc = GGI_LL_MATHGC(buf);
	if (mathgc == NULL) goto inval;

	mathmask = 0;

	/* Caveat: below we take advantage that the LL math enum and the 
	 * LibBuf flags are mostly identical.  If this fact changes, we
	 * will need to convert here.
	 */
	switch(ggiGAGetType(tank_h)) {
	case GA_RT_BUFFER_ZBUFFER:
	  mathmask = BUF_CMP_MASK | BUF_CMP_SIGNED;
	  *flags = mathgc->math & mathmask;
	  
	  if (mathgc->math & LL_MATH_Z_DISPOSE) *flags |= BUF_DISPOSE;
	  if (mathgc->math & (LL_MATH_Z_DISPOSE << LL_MATH_ALTSHIFT))
	    *flags |= BUF_DISPOSE << BUF_TOGGLE_SHIFT;
	  break;
	case GA_RT_BUFFER_ABUFFER:
	  mathmask = BUF_BLENDTYPE_MASK | BUF_CMP_MASK | BUF_CMP_SIGNED |
	    (BUF_BLENDTYPE_MASK << BUF_ALTBLEND_SHIFT) | BUF_DISPOSE |
	    (BUF_DISPOSE << BUF_TOGGLE_SHIFT);
	  *flags = mathgc->math & mathmask;
	  break;
	default:
	  BUFDPRINT("Buffer not Alpha or Z, getflags not implemented yet.\n");
	  return GGI_EARGINVAL;
	  break;
	}
	return 0;
 inval:
	BUFDPRINT("Fishy buf_t in ggiBufGetGCFlags\n");
	return GGI_EARGINVAL;
}	/* ggiBufGetGCFlags */

int ggiBufClearBuffer(ggiBuf_t buf)
{
	ggi_ll_fgbg_gc_t gc;
	ggi_ll_ststate_t ststate;

	if (buf == NULL) return GGI_EARGINVAL;

	gc = GGI_LL_GC(buf);
	if (gc == NULL) return GGI_EARGINVAL;
	ststate = GGI_LL_STSTATE(buf);
	if (ststate == NULL) return GGI_EARGINVAL;

	return ggiBufConst2Buf(buf,	        /* dest */
			       0,		/* xoffset */
			       0,		/* yoffset */
			       0,		/* stride */
			       gc->bgsrc,	/* bg_val */
			       ststate->w, ststate->h);
}	/* ggiBufClearBuffer */

int ggiBufGetValue(ggiBuf_t buf, size_t xoffset, size_t yoffset,
		   size_t stride, ggi_bufval *val)
{
	return ggiBufBuf2Mem(buf, xoffset, yoffset, stride,
			     val, sizeof(*val), sizeof(*val), 1);
}	/* ggiBufGetValue */

int ggiBufPutValue(ggiBuf_t buf, size_t xoffset, size_t yoffset,
		   size_t stride, ggi_bufval val)
{
	return ggiBufMem2Buf(buf, xoffset, yoffset, stride,
			     &val, sizeof(val), sizeof(val), 1);
}	/* ggiBufPutValue */
